:root {
    --parsl-dark-blue: #2D5992;
    --parsl-light-blue: #83B0DC;
}

table {
    border-collapse: collapse;
}
th {
    background: #ccc;
    padding: 8px;
}

tr {
    height: 30px;
    line-height: 30px;
}

th, td {
    border: 1px solid #ccc;
}

tr:nth-child(even) {
    background: #efefef;
}

tr:hover {
    background: #d1d1d1;
}

/* Deal with td expansion and overflow. Contain table to window. Table scrolliing */
td {
    width: 250px;
    height: 100%;
    white-space: nowrap;
    text-align: center;
}

td a {
    box-sizing: border-box;
    display: inline-block;
    height:100%;
    width:100%;
    padding: 8px;
    color: black;
    text-decoration: none;
    text-overflow: ellipsis;
}

#workflows_table tr > *:nth-child(1) {
    display: none;
}